"""Processing of ULVAC PHI Multipak XPS traces.

The [IGOR `.spe` import script by jjweimer](https://www.wavemetrics.com/project/phispefileloader)
was pretty helpful for writing this parser.

File Structure of `.spe` Files
``````````````````````````````

These binary files actually contain an ASCII file header, delimited by
`"SOFH\n"` and `"EOFH\n"`.

The binding energies corresponding to the datapoints in the later part
of the file can be found from the `"SpectralRegDef"` entries in this
header. Each of these entries look something like:

.. code-block::

    2 2 F1s 9 161 -0.1250 695.0 675.0 695.0 680.0    0.160000 29.35 AREA

This maps as follows:

.. code-block::

    2           trace_number
    2           trace_number (again?)
    F1s         name
    9           atomic_number
    161         num_datapoints
    -0.1250     step
    695.0       start
    675.0       stop
    695.0       ?
    680.0       ?
    0.160000    dwell_time
    29.35       e_pass
    AREA        description (?)

After the file header, the binary part starts with a short data header
(offsets given from start of data header):

.. code-block::

    0x0000 group                # Data group number.
    0x0004 num_traces           # Number of traces in file
    0x0008 trace_header_size    # Combined lengths of all trace headers.
    0x000c data_header_size     # Length of this data header.

After this follow `num_traces` trace headers that are each structured
something like this:

.. code-block::

    0x0000 trace_number          # Number of the trace.
    0x0004 bool_01               # ???
    0x0008 bool_02               # ???
    0x000c trace_number_again    # Number of the trace. Again?
    0x0010 bool_03               # ???
    0x0014 num_datapoints        # Number of datapoints in trace.
    0x0018 bool_04               # ???
    0x001c bool_05               # ???
    0x0020 string_01             # ???
    0x0024 string_02             # ???
    0x0028 string_03             # ???
    0x002c int_02                # ???
    0x0030 string_04             # ???
    0x0034 string_05             # ???
    0x0038 y_unit                # The unit of the datapoints.
    0x003c int_05                # ???
    0x0040 int_06                # ???
    0x0044 int_07                # ???
    0x0048 data_dtype            # Data type for datapoints (`f4`/`f8`).
    0x004c num_data_bytes        # Unsure about this one.
    0x0050 num_datapoints_tot    # This one as well.
    0x0054 int_10                # ???
    0x0058 int_11                # ???
    0x005c end_of_data           # Byte offset of the end-of-data.

After the trace headers follow the datapoints. After the number of
datapoints there is a single 32bit float with the trace's dwelling time
again.

Structure of Parsed Data
````````````````````````

.. code-block:: yaml

    - fn   !!str
    - uts  !!float
    - raw:
        "{{ trace_number }}":
          {n: [!!float, ...], s: [!!float, ...], u: !!str}

.. codeauthor:: Nicolas Vetsch <vetschnicolas@gmail.com>
"""

import math
import os
import re
from typing import Any

import numpy as np

data_header_dtype = np.dtype(
    [
        ("group", "<u4"),
        ("num_traces", "<u4"),
        ("trace_header_size", "<u4"),
        ("data_header_size", "<u4"),
    ]
)

trace_header_dtype = np.dtype(
    [
        ("trace_number", "<u4"),
        ("bool_01", "<u4"),
        ("bool_02", "<u4"),
        ("trace_number_again", "<u4"),
        ("bool_03", "<u4"),
        ("num_datapoints", "<u4"),
        ("bool_04", "<u4"),
        ("bool_05", "<u4"),
        ("string_01", "|S4"),  # pnt?
        ("string_02", "|S4"),
        ("string_03", "|S4"),  # sar?
        ("int_02", "<u4"),
        ("string_04", "|S4"),
        ("string_05", "|S4"),
        ("y_unit", "|S4"),
        ("int_05", "<u4"),
        ("int_06", "<u4"),
        ("int_07", "<u4"),
        ("data_dtype", "|S4"),
        ("num_data_bytes", "<u4"),
        ("num_datapoints_tot", "<u4"),
        ("int_10", "<u4"),
        ("int_11", "<u4"),
        ("end_of_data", "<u4"),
    ]
)


def camel_to_snake(s: str) -> str:
    """Converts CamelCase strings to snake_case.

    # From https://stackoverflow.com/a/1176023

    Parameters
    ----------
    s
        The CamelCase input string.

    Returns
    -------
    str
        The snake_case equivalent of s.

    """
    s = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def _read_pascal_string(pascal_bytes: bytes, encoding: str = "utf-8") -> str:
    """Parses a length-prefixed string given some encoding.

    Parameters
    ----------
    bytes
        The bytes of the string starting at the length-prefix byte.

    encoding
        The encoding of the string to be converted.

    Returns
    -------
    str
        The string decoded from the input bytes.

    """
    if len(pascal_bytes) < pascal_bytes[0] + 1:
        raise ValueError("Insufficient number of bytes.")
    string_bytes = pascal_bytes[1 : pascal_bytes[0] + 1]
    return string_bytes.decode(encoding)


def _read_value(
    data: bytes, offset: int, dtype: np.dtype, encoding: str = "utf-8"
) -> Any:
    """Reads a single value from a buffer at a certain offset.

    Just a handy wrapper for np.frombuffer() With the added benefit of
    allowing the 'pascal' keyword as an indicator for a length-prefixed
    string.

    The read value is converted to a built-in datatype using
    np.dtype.item().

    Parameters
    ----------
    data
        An object that exposes the buffer interface. Here always bytes.

    offset
        Start reading the buffer from this offset (in bytes).

    dtype
        Data-type to read in.

    encoding
        The encoding of the bytes to be converted.

    Returns
    -------
    Any
        The unpacked and converted value from the buffer.

    """
    if dtype == "pascal":
        # Allow the use of 'pascal' in all of the dtype maps.
        return _read_pascal_string(data[offset:])
    value = np.frombuffer(data, offset=offset, dtype=dtype, count=1)
    item = value.item()
    if value.dtype.names:
        item = [i.decode(encoding) if isinstance(i, bytes) else i for i in item]
        return dict(zip(value.dtype.names, item))
    return item.decode(encoding) if isinstance(item, bytes) else item


def _read_values(data: bytes, offset: int, dtype, count) -> list:
    """Reads in multiple values from a buffer starting at offset.

    Just a handy wrapper for np.frombuffer() with count >= 1.

    The read values are converted to a list of built-in datatypes using
    np.ndarray.tolist().

    Parameters
    ----------
    data
        An object that exposes the buffer interface. Here always bytes.

    offset
        Start reading the buffer from this offset (in bytes).

    dtype
        Data-type to read in.

    count
        Number of items to read. -1 means all data in the buffer.

    Returns
    -------
    Any
        The values read from the buffer as specified by the arguments.

    """
    values = np.frombuffer(data, offset=offset, dtype=dtype, count=count)
    if values.dtype.names:
        return [dict(zip(value.dtype.names, value.item())) for value in values]
    # The ndarray.tolist() method converts numpy scalars to built-in
    # scalars, hence not just list(values).
    return values.tolist()


def _process_header(spe: list[bytes]) -> dict:
    """Processes the file header at the top of `.spe` files.

    Parameters
    ----------
    spe
        The lines of bytes read from file.

    Returns
    -------
    dict
        The file header parsed into a dictionary. Some entries (keys)
        occur more than once. The corresponding values are joined into
        a list.

    """
    header_lines = spe[spe.index(b"SOFH\n") + 1 : spe.index(b"EOFH\n")]
    header = {}
    for line in header_lines:
        key, value = line.split(b":")
        key, value = camel_to_snake(key.decode().strip()), value.decode().strip()
        if key in header:
            header[key] = [header[key]] + [value]
        else:
            header[key] = value
    return header


def _process_trace_defs(header: dict) -> list[dict]:
    """Processes the trace definition strings given in the file header.

    These strings look something like the following:
    `2 2 F1s 9 161 -0.1250 695.0 675.0 695.0 680.0    0.160000 29.35 AREA`

    Parameters
    ----------
    header
        The file header parsed into a dictionary. The "SpectralRegDef"
        entry contains a list of trace definition strings.

    Returns
    -------
    list[dict]
        A list of trace definition dictionaries describind the kind of
        trace and the binding energy ranges.

    """
    trace_defs = []
    for trace_def in header.get("spectral_reg_def"):
        trace_def = trace_def.split()
        trace_defs.append(
            {
                "trace_number": int(trace_def[0]),
                "name": trace_def[2],
                "atomic_number": trace_def[3],
                "num_datapoints": int(trace_def[4]),
                "step": float(trace_def[5]),
                "start": float(trace_def[6]),
                "stop": float(trace_def[7]),
                "dwell_time": trace_def[10],
                "e_pass": trace_def[11],
                "description": trace_def[12],
            }
        )
    return trace_defs


def _process_traces(spe: list[bytes], trace_defs: list[dict]) -> dict:
    """Processes the spectral traces in the file.

    Parameters
    ----------
    spe
        The lines of bytes read from file.

    trace_defs
        The list of trace definitions parsed from the file header.

    Returns
    -------
    dict
        A dictionary containing the binding energies constructed from
        the trace definitions and the corrresponding XPS traces.

    """
    data = b"".join(spe[spe.index(b"EOFH\n") + 1 :])
    data_header = _read_value(data, 0x0000, data_header_dtype)
    assert data_header["num_traces"] == len(trace_defs)
    # All trace headers I have seen are 192 (0xc0) bytes long.
    assert data_header["trace_header_size"] / trace_header_dtype.itemsize == len(
        trace_defs
    )
    assert data_header["data_header_size"] == data_header_dtype.itemsize
    trace_headers = _read_values(data, 0x0010, trace_header_dtype, len(trace_defs))
    traces = {}
    for trace_header, trace_def in zip(trace_headers, trace_defs):
        assert trace_header["trace_number"] == trace_def["trace_number"]
        assert trace_header["num_datapoints"] == trace_def["num_datapoints"]
        # Contruct the binding energies from trace_def.
        energies, dE = np.linspace(
            trace_def["start"],
            trace_def["stop"],
            trace_def["num_datapoints"],
            endpoint=True,
            retstep=True,
        )
        E = {"n": energies.tolist(), "s": [abs(dE)] * len(energies), "u": "eV"}
        # Construct data from trace_header
        data_dtype = np.dtype(f'{trace_header["data_dtype"].decode()}')
        data_offset = trace_header["end_of_data"] - trace_header["num_data_bytes"]
        datapoints = _read_values(
            data, data_offset, data_dtype, trace_header["num_datapoints"]
        )
        dwell_time = _read_value(data, trace_header["end_of_data"], "<f4")
        np.testing.assert_almost_equal(dwell_time, float(trace_def["dwell_time"]))
        # TODO: Figure out the correct error. This signal count should
        # somehow be a Poisson distribution, i.e. the error should be
        # something like s ~ sqrt(n). The counts per second only seem to
        # be taking values in steps of 12.5cps, so taking this as error.
        y = {
            "n": datapoints,
            "s": [12.5] * len(datapoints),
            "u": "cps",
        }
        traces[str(trace_def["trace_number"])] = {
            "name": trace_def["name"],
            "atomic_number": trace_def["atomic_number"],
            "dwell_time": trace_def["dwell_time"],
            "e_pass": trace_def["e_pass"],
            "description": trace_def["description"],
            "E": E,
            "y": y,
        }
    return traces


def process(
    fn: str, encoding: str = "utf-8", timezone: str = "UTC"
) -> tuple[list, dict, bool]:
    """Processes ULVAC-PHI Multipak XPS data.

    Parameters
    ----------
    fn
        The file containing the data to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "UTC".

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and the full date tag.
        Multipak .spe files seemingly have no timestamp.

    """
    with open(fn, "rb") as spe_file:
        spe = spe_file.readlines()
    header = _process_header(spe)
    software_id, version = header.get("software_version").split()
    meta = {
        "params": {
            "software_id": software_id,
            "version": version,
            "username": header.get("operator"),
        },
        "file_header": header,
    }
    trace_defs = _process_trace_defs(header)
    traces = _process_traces(spe, trace_defs)
    timesteps = [{"fn": fn, "raw": {"traces": traces}}]
    return timesteps, meta, False



"""
File parser for Agilent OpenLab binary signal trace files (CH and IT).

Currently supports version "179" of the files. Version information is defined in
the `magic_values` (parameters & metadata) and `data_dtypes` (data) dictionaries.

Adapted from `ImportAgilent.m <https://bit.ly/3HSelIR>`_ and 
`aston <https://github.com/bovee/Aston>`_.

Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      sampleid: !!str
      username: !!str
      version:  !!str
      valve:    None
      datafile: None

File Structure of ``.ch`` files
```````````````````````````````
.. code ::

    0x0000 "version magic"
    0x0108 "data offset"
    0x011a "x-axis minimum (ms)"
    0x011e "x-axis maximum (ms)"
    0x035a "sample ID"
    0x0559 "description"
    0x0758 "username"
    0x0957 "timestamp"
    0x09e5 "instrument name"
    0x09bc "inlet"
    0x0a0e "method"
    0x104c "y-axis unit"
    0x1075 "detector name"
    0x1274 "y-axis intercept"
    0x127c "y-axis slope"

Data is stored in a consecutive set of ``<f8``, starting at the offset (calculated
as ``offset =  ("data offset" - 1) * 512``) until the end of the file.

.. codeauthor:: Peter Kraus <peter.kraus@empa.ch>
"""
import numpy as np

import yadg.dgutils
from yadg.dgutils.dateutils import str_to_uts

magic_values = {}
magic_values["179"] = {
    0x035A: ("sampleid", "utf-16"),
    0x0559: ("description", "utf-16"),
    0x0A0E: ("method", "utf-16"),
    0x0758: ("username", "utf-16"),
    0x0957: ("timestamp", "utf-16"),
    0x09E5: ("instrument", "utf-16"),
    0x09BC: ("inlet", "utf-16"),
    0x104C: ("yunit", "utf-16"),
    0x1075: ("tracetitle", "utf-16"),
    0x0108: ("offset", ">i4"),  # (x-1) * 512
    0x011A: ("xmin", ">f4"),  # / 60000
    0x011E: ("xmax", ">f4"),  # / 60000
    0x1274: ("intercept", ">f8"),
    0x127C: ("slope", ">f8"),
}

data_dtypes = {}
data_dtypes["179"] = (8, "<f8")


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict]:
    """
    Agilent OpenLAB signal trace parser

    One chromatogram per file with a single trace. Binary data format.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Not used as the file is binary.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    ([chrom], metadata): tuple[list, dict]
        Standard timesteps & metadata tuple.
    """

    with open(fn, "rb") as infile:
        magic = yadg.dgutils.read_value(infile, 0, "utf-8", 1)
        pars = {}
        if magic in magic_values.keys():
            for offset, (tag, dtype) in magic_values[magic].items():
                v = yadg.dgutils.read_value(infile, offset, dtype, 1)
                pars[tag] = v
        pars["end"] = infile.seek(0, 2)
    dsize, ddtype = data_dtypes[magic]
    pars["start"] = (pars["offset"] - 1) * 512
    nbytes = pars["end"] - pars["start"]
    assert nbytes % dsize == 0
    npoints = nbytes // dsize

    xsn = np.linspace(pars["xmin"] / 1000, pars["xmax"] / 1000, num=npoints)
    xss = np.ones(npoints) * xsn[0]
    with open(fn, "rb") as infile:
        ysn = (
            yadg.dgutils.read_value(infile, pars["start"], ddtype, npoints)
            * pars["slope"]
        )
    yss = np.ones(npoints) * pars["slope"]

    detector, title = pars["tracetitle"].split(",")

    chrom = {
        "fn": str(fn),
        "uts": str_to_uts(pars["timestamp"], "%d-%b-%y, %H:%M:%S", timezone),
        "traces": {
            detector: {
                "t": {"n": xsn.tolist(), "s": xss.tolist(), "u": "s"},
                "y": {"n": ysn.tolist(), "s": yss.tolist(), "u": pars["yunit"]},
                "id": 0,
                "data": [(xsn, xss), (ysn, yss)],
            }
        },
    }
    metadata = {"type": "agilent.ch", "params": {}}

    for k in ["sampleid", "username", "method"]:
        metadata["params"][k] = pars[k]
    metadata["params"]["valve"] = None
    metadata["params"]["version"] = str(magic)
    metadata["params"]["datafile"] = None

    return [chrom], metadata



import os
import logging
from scipy.signal import find_peaks
import numpy as np

import uncertainties as uc
from uncertainties import unumpy as unp


import yadg.dgutils
from yadg.parsers.qftrace import fit, prune, labviewcsv

version = "4.0.0"


def _fit(
    fvals: np.ndarray,
    fsigs: np.ndarray,
    gvals: np.ndarray,
    absgvals: np.ndarray,
    method: str,
    height: float,
    distance: float,
    cutoff: float,
    threshold: float,
) -> tuple[dict]:
    """
    Function that fits Q and f from Γ(f).
    """
    Qs = {"n": [], "s": [], "u": " "}
    fs = {"n": [], "s": [], "u": "Hz"}
    mag = -10 * np.log10(absgvals)
    peaks, _ = find_peaks(mag, height=height, distance=distance)
    assert len(peaks) > 0, logging.error("qftrace: No peaks were found.")
    npeaks = len(peaks)
    if method == "lorentz":
        _prune = prune.gradient
        _fitq = fit.lorentz
        _ppar = threshold
    elif method == "kajfez" or method == "q0refl":
        _prune = prune.cutoff
        _fitq = fit.kajfez
        _ppar = cutoff
    elif method == "naive":
        _prune = prune.gradient
        _fitq = fit.naive
        _ppar = threshold
    for p in peaks:
        # prune using a smaller number of points
        n = gvals.size
        os = max(0, p - max(1000, n // 5))
        oe = min(n, p + max(1000, n // 5))
        s, e = _prune(p - os, absgvals[os:oe], _ppar)
        s = os + s
        e = os + e
        Q, f = _fitq(fvals[s:e], fsigs[s:e], gvals[s:e], absgvals[s:e])
        Qs["n"].append(Q.n)
        Qs["s"].append(Q.s)
        fs["n"].append(f.n)
        fs["s"].append(f.s)
    assert len(Qs["n"]) == npeaks and len(fs["n"]) == npeaks
    return Qs, fs


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "timezone",
    tracetype: str = "labview.csv",
    method: str = "kajfez",
    height: float = 1.0,
    distance: float = 5000.0,
    cutoff: float = 0.4,
    threshold: float = 1e-6,
) -> tuple[list, dict, bool]:
    """
    VNA reflection trace parser.

    This parser processes a VNA log file, containing the complex reflection coefficient
    data as a function of frequency (:math:`\\Gamma(f)`). This data is automatically
    worked up to produce the quality factor :math:`Q_0` and the central frequency
    :math:`f_0` of all peaks found in each trace; hence the name ``qftrace``.

    Parameters
    ----------
    fn
        File to process

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "localtime".

    method
        Method used for fitting :math:`Q_0` and :math:`f_0` to :math:`\\Gamma(f)` data.
        One of ``"naive"``, ``"lorentz"``, or ``"kajfez"``. Default is ``"kajfez"``.

    cutoff
        Parameter passed to the cutoff-based pruning routine, defining the cutoff
        threshold for the normalisation. Default is 0.4.

    threshold
        Parameter passed to the gradient-based pruning routine, defining the minimum
        gradient below which the trace gets pruned. Default is 1e-6.

    height
        Parameter for the peak-picker.

    distance
        Parameter for the peak-picker.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and full date tag. The currently only
        supported tracetype ("labview.csv") does not return full date.
    """
    if tracetype == "labview.csv":
        data, meta = labviewcsv.process(fn, encoding, timezone)
        fulldate = False
    for ts in data:
        for det, trace in ts["raw"]["traces"].items():
            Q, f = _fit(
                trace.pop("fvals"),
                trace.pop("fsigs"),
                trace.pop("gamma"),
                trace.pop("absgamma"),
                method,
                height,
                distance,
                cutoff,
                threshold,
            )
            if "derived" not in ts:
                ts["derived"] = {}
            ts["derived"][det] = {"Q": Q, "f": f}
    return data, meta, fulldate



"""Processing of BioLogic's EC-Lab binary modular files.

This code is partly an adaptation of the [`galvani` module by Chris
Kerr](https://github.com/echemdata/galvani) and builds on the work done
by the previous civilian service member working on the project, Jonas
Krieger.

File Structure of `.mpr` Files
``````````````````````````````

At a top level, `.mpr` files are made up of a number of modules,
separated by the `MODULE` keyword. In all the files I have seen, the
first module is the settings module, followed by the data module, the
log module and then an optional loop module.

.. code-block::

    0x0000 BIO-LOGIC MODULAR FILE  # File magic.
    0x0034 MODULE                  # Module magic.
    ...                            # Module 1.
    0x???? MODULE                  # Module magic.
    ...                            # Module 2.
    0x???? MODULE                  # Module magic.
    ...                            # Module 3.
    0x???? MODULE                  # Module magic.
    ...                            # Module 4.

After splitting the entire file on `MODULE`, each module starts with a
header that is structured like this (offsets from start of module):

.. code-block::

    0x0000 short_name  # Short name, e.g. VMP Set.
    0x000A long_name   # Longer name, e.g. VMP settings.
    0x0023 length      # Number of bytes in module data.
    0x0027 version     # Module version.
    0x002B date        # Acquisition date in ASCII, e.g. 08/10/21.
    ...                # Module data.

The contents of each module's data vary wildly depending on the used
technique, the module and perhaps the software version, the settings in
EC-Lab, etc. Here a quick overview (offsets from start of module data).

*Settings Module*

.. code-block::

    0x0000 technique_id           # Unique technique ID.
    ...                           # ???
    0x0007 comments               # Pascal string.
    ...                           # Zero padding.
    # Cell Characteristics.
    0x0107 active_material_mass   # Mass of active material
    0x010B at_x                   # at x =
    0x010F molecular_weight       # Molecular weight of active material
    0x0113 atomic_weight          # Atomic weight of intercalated ion
    0x0117 acquisition_start      # Acquisition started a: xo =
    0x011B e_transferred          # Number of e- transferred
    0x011E electrode_material     # Pascal string.
    ...                           # Zero Padding
    0x01C0 electrolyte            # Pascal string.
    ...                           # Zero Padding, ???.
    0x0211 electrode_area         # Electrode surface area
    0x0215 reference_electrode    # Pascal string
    ...                           # Zero padding
    0x024C characteristic_mass    # Characteristic mass
    ...                           # ???
    0x025C battery_capacity       # Battery capacity C =
    0x0260 battery_capacity_unit  # Unit of the battery capacity.
    ...                           # ???
    # Technique parameters can randomly be found at 0x0572, 0x1845 or
    # 0x1846. All you can do is guess and try until it fits.
    0x1845 ns                     # Number of sequences.
    0x1847 n_params               # Number of technique parameters.
    0x1849 params                 # ns sets of n_params parameters.
    ...                           # ???

*Data Module*

.. code-block::

    0x0000 n_datapoints  # Number of datapoints.
    0x0004 n_columns     # Number of values per datapoint.
    0x0005 column_ids    # n_columns unique column IDs.
    ...
    # Depending on module version, datapoints start 0x0195 or 0x0196.
    # Length of each datapoint depends on number and IDs of columns.
    0x0195 datapoints    # n_datapoints points of data.

*Log Module*

.. code-block::

    ...                         # ???
    0x0009 channel_number       # Zero-based channel number.
    ...                         # ???
    0x00AB channel_sn           # Channel serial number.
    ...                         # ???
    0x01F8 Ewe_ctrl_min         # Ewe ctrl range min.
    0x01FC Ewe_ctrl_max         # Ewe ctrl range max.
    ...                         # ???
    0x0249 ole_timestamp        # Timestamp in OLE format.
    0x0251 filename             # Pascal String.
    ...                         # Zero padding, ???.
    0x0351 host                 # IP address of host, Pascal string.
    ...                         # Zero padding.
    0x0384 address              # IP address / COM port of potentiostat.
    ...                         # Zero padding.
    0x03B7 ec_lab_version       # EC-Lab version (software)
    ...                         # Zero padding.
    0x03BE server_version       # Internet server version (firmware)
    ...                         # Zero padding.
    0x03C5 interpreter_version  # Command interpretor version (firmware)
    ...                         # Zero padding.
    0x03CF device_sn            # Device serial number.
    ...                         # Zero padding.
    0x0922 averaging_points     # Smooth data on ... points.
    ...                         # ???

*Loop Module*

.. code-block::

    0x0000 n_indexes  # Number of loop indexes.
    0x0004 indexes    # n_indexes indexes at which loops start in data.
    ...               # ???


Structure of Parsed Data
````````````````````````

*EIS Techniques (PEIS/GEIS)*

.. code-block:: yaml

    - fn   !!str
    - uts  !!float
    - raw:
        traces:
          "{{ technique name }}":
            "{{ col1 }}":
              [!!int, ...]
            "{{ col2 }}":
              {n: [!!float, ...], s: [!!float, ...], u: !!str}

*All Other Techniques*

.. code-block:: yaml

    - fn   !!str
    - uts  !!float
    - raw:
        "{{ col1 }}":  !!int
        "{{ col2 }}":
          {n: !!float, s: !!float, u: !!str}

.. codeauthor:: Nicolas Vetsch <vetschnicolas@gmail.com>
"""
import logging
import math
from collections import defaultdict
from typing import Any

import numpy as np
from yadg.dgutils.dateutils import ole_to_uts
from yadg.parsers.electrochem.eclabtechniques import technique_params_dtypes
from yadg.parsers.electrochem.eclabtechniques import param_from_key, get_resolution

# Module header starting after each MODULE keyword.
module_header_dtype = np.dtype(
    [
        ("short_name", "|S10"),
        ("long_name", "|S25"),
        ("length", "<u4"),
        ("version", "<u4"),
        ("date", "|S8"),
    ]
)


# Relates the offset in the settings data to the corresponding dtype and
# name. Maybe watch out for very long pascal strings as they may span
# over the entire zero padding and shift the offsets.
settings_dtypes = {
    0x0007: ("pascal", "comments"),
    0x0107: ("<f4", "active_material_mass"),
    0x010B: ("<f4", "at_x"),
    0x010F: ("<f4", "molecular_weight"),
    0x0113: ("<f4", "atomic_weight"),
    0x0117: ("<f4", "acquisition_start"),
    0x011B: ("<u2", "e_transferred"),
    0x011E: ("pascal", "electrode_material"),
    0x01C0: ("pascal", "electrolyte"),
    0x0211: ("<f4", "electrode_area"),
    0x0215: ("pascal", "reference_electrode"),
    0x024C: ("<f4", "characteristic_mass"),
    0x025C: ("<f4", "battery_capacity"),
    0x0260: ("|u1", "battery_capacity_unit"),
    # NOTE: The compliance limits are apparently not always at this
    # offset, hence commented out...
    # 0x19d2: ("<f4", "compliance_min"),
    # 0x19d6: ("<f4", "compliance_max"),
}


# Maps the flag column ID bytes to the corresponding bitmask and name.
flag_columns = {
    0x0001: (0b00000011, "mode"),
    0x0002: (0b00000100, "ox/red"),
    0x0003: (0b00001000, "error"),
    0x0015: (0b00010000, "control changes"),
    0x001F: (0b00100000, "Ns changes"),
    # NOTE: I think the missing bitmask (0b01000000) is a stop bit. It
    # sometimes appears in the flag byte of the very last data point.
    0x0041: (0b10000000, "counter inc."),
}


# Maps the data column ID bytes to the corresponding dtype, name and
# unit.
data_columns = {
    0x0004: ("<f8", "time", "s"),
    0x0005: ("<f4", "control_V/I", "V/mA"),
    0x0006: ("<f4", "Ewe", "V"),
    0x0007: ("<f8", "dq", "mA·h"),
    0x0008: ("<f4", "I", "mA"),
    0x0009: ("<f4", "Ece", "V"),
    0x000B: ("<f8", "<I>", "mA"),
    0x000D: ("<f8", "(Q-Qo)", "mA·h"),
    0x0010: ("<f4", "Analog IN 1", "V"),
    0x0011: ("<f4", "Analog IN 2", "V"),
    0x0013: ("<f4", "control_V", "V"),
    0x0014: ("<f4", "control_I", "mA"),
    0x0017: ("<f8", "dQ", "mA·h"),
    0x0018: ("<f8", "cycle number", None),
    0x0020: ("<f4", "freq", "Hz"),
    0x0021: ("<f4", "|Ewe|", "V"),
    0x0022: ("<f4", "|I|", "A"),
    0x0023: ("<f4", "Phase(Z)", "deg"),
    0x0024: ("<f4", "|Z|", "Ω"),
    0x0025: ("<f4", "Re(Z)", "Ω"),
    0x0026: ("<f4", "-Im(Z)", "Ω"),
    0x0027: ("<u2", "I Range", None),
    0x0046: ("<f4", "P", "W"),
    0x004A: ("<f8", "Energy", "W·h"),
    0x004B: ("<f4", "Analog OUT", "V"),
    0x004C: ("<f4", "<I>", "mA"),
    0x004D: ("<f4", "<Ewe>", "V"),
    0x004E: ("<f4", "Cs⁻²", "µF⁻²"),
    0x0060: ("<f4", "|Ece|", "V"),
    0x0062: ("<f4", "Phase(Zce)", "deg"),
    0x0063: ("<f4", "|Zce|", "Ω"),
    0x0064: ("<f4", "Re(Zce)", "Ω"),
    0x0065: ("<f4", "-Im(Zce)", "Ω"),
    0x007B: ("<f8", "Energy charge", "W·h"),
    0x007C: ("<f8", "Energy discharge", "W·h"),
    0x007D: ("<f8", "Capacitance charge", "µF"),
    0x007E: ("<f8", "Capacitance discharge", "µF"),
    0x0083: ("<u2", "Ns", None),
    0x00A3: ("<f4", "|Estack|", "V"),
    0x00A8: ("<f4", "Rcmp", "Ω"),
    0x00A9: ("<f4", "Cs", "µF"),
    0x00AC: ("<f4", "Cp", "µF"),
    0x00AD: ("<f4", "Cp⁻²", "µF⁻²"),
    0x00AE: ("<f4", "<Ewe>", "V"),
    0x00F1: ("<f4", "|E1|", "V"),
    0x00F2: ("<f4", "|E2|", "V"),
    0x010F: ("<f4", "Phase(Z1)", "deg"),
    0x0110: ("<f4", "Phase(Z2)", "deg"),
    0x012D: ("<f4", "|Z1|", "Ω"),
    0x012E: ("<f4", "|Z2|", "Ω"),
    0x014B: ("<f4", "Re(Z1)", "Ω"),
    0x014C: ("<f4", "Re(Z2)", "Ω"),
    0x0169: ("<f4", "-Im(Z1)", "Ω"),
    0x016A: ("<f4", "-Im(Z2)", "Ω"),
    0x0187: ("<f4", "<E1>", "V"),
    0x0188: ("<f4", "<E2>", "V"),
    0x01A6: ("<f4", "Phase(Zstack)", "deg"),
    0x01A7: ("<f4", "|Zstack|", "Ω"),
    0x01A8: ("<f4", "Re(Zstack)", "Ω"),
    0x01A9: ("<f4", "-Im(Zstack)", "Ω"),
    0x01AA: ("<f4", "<Estack>", "V"),
    0x01AE: ("<f4", "Phase(Zwe-ce)", "deg"),
    0x01AF: ("<f4", "|Zwe-ce|", "Ω"),
    0x01B0: ("<f4", "Re(Zwe-ce)", "Ω"),
    0x01B1: ("<f4", "-Im(Zwe-ce)", "Ω"),
    0x01B2: ("<f4", "(Q-Qo)", "C"),
    0x01B3: ("<f4", "dQ", "C"),
    0x01B9: ("<f4", "<Ece>", "V"),
    0x01CE: ("<f4", "Temperature", "°C"),
    0x01D3: ("<f8", "Q charge/discharge", "mA·h"),
    0x01D4: ("<u4", "half cycle", None),
    0x01D5: ("<u4", "z cycle", None),
    0x01D7: ("<f4", "<Ece>", "V"),
    0x01D9: ("<f4", "THD Ewe", "%"),
    0x01DA: ("<f4", "THD I", "%"),
    0x01DC: ("<f4", "NSD Ewe", "%"),
    0x01DD: ("<f4", "NSD I", "%"),
    0x01DF: ("<f4", "NSR Ewe", "%"),
    0x01E0: ("<f4", "NSR I", "%"),
    0x01E6: ("<f4", "|Ewe h2|", "V"),
    0x01E7: ("<f4", "|Ewe h3|", "V"),
    0x01E8: ("<f4", "|Ewe h4|", "V"),
    0x01E9: ("<f4", "|Ewe h5|", "V"),
    0x01EA: ("<f4", "|Ewe h6|", "V"),
    0x01EB: ("<f4", "|Ewe h7|", "V"),
    0x01EC: ("<f4", "|I h2|", "A"),
    0x01ED: ("<f4", "|I h3|", "A"),
    0x01EE: ("<f4", "|I h4|", "A"),
    0x01EF: ("<f4", "|I h5|", "A"),
    0x01F0: ("<f4", "|I h6|", "A"),
    0x01F1: ("<f4", "|I h7|", "A"),
}


# Relates the offset in log data to the corresponding dtype and name.
# NOTE: The safety limits are maybe at 0x200?
# NOTE: The log also seems to contain the settings again. These are left
# away for now.
# NOTE: Looking at .mpl files, the log module appears to consist of
# multiple 'modify on' sections, each starting with an OLE timestamp.
log_dtypes = {
    0x0009: ("|u1", "channel_number"),
    0x00AB: ("<u2", "channel_sn"),
    0x01F8: ("<f4", "Ewe_ctrl_min"),
    0x01FC: ("<f4", "Ewe_ctrl_max"),
    0x0249: ("<f8", "ole_timestamp"),
    0x0251: ("pascal", "filename"),
    0x0351: ("pascal", "host"),
    0x0384: ("pascal", "address"),
    0x03B7: ("pascal", "ec_lab_version"),
    0x03BE: ("pascal", "server_version"),
    0x03C5: ("pascal", "interpreter_version"),
    0x03CF: ("pascal", "device_sn"),
    0x0922: ("|u1", "averaging_points"),
}


def _read_pascal_string(pascal_bytes: bytes, encoding: str = "windows-1252") -> str:
    """Parses a length-prefixed string given some encoding.

    Parameters
    ----------
    bytes
        The bytes of the string starting at the length-prefix byte.

    encoding
        The encoding of the string to be converted.

    Returns
    -------
    str
        The string decoded from the input bytes.

    """
    if len(pascal_bytes) < pascal_bytes[0] + 1:
        raise ValueError("Insufficient number of bytes.")
    string_bytes = pascal_bytes[1 : pascal_bytes[0] + 1]
    return string_bytes.decode(encoding)


def _read_value(
    data: bytes, offset: int, dtype: np.dtype, encoding: str = "windows-1252"
) -> Any:
    """Reads a single value from a buffer at a certain offset.

    Just a handy wrapper for np.frombuffer() With the added benefit of
    allowing the 'pascal' keyword as an indicator for a length-prefixed
    string.

    The read value is converted to a built-in datatype using
    np.dtype.item().

    Parameters
    ----------
    data
        An object that exposes the buffer interface. Here always bytes.

    offset
        Start reading the buffer from this offset (in bytes).

    dtype
        Data-type to read in.

    encoding
        The encoding of the bytes to be converted.

    Returns
    -------
    Any
        The unpacked and converted value from the buffer.

    """
    if dtype == "pascal":
        # Allow the use of 'pascal' in all of the dtype maps.
        return _read_pascal_string(data[offset:])
    value = np.frombuffer(data, offset=offset, dtype=dtype, count=1)
    item = value.item()
    if value.dtype.names:
        item = [i.decode(encoding) if isinstance(i, bytes) else i for i in item]
        return dict(zip(value.dtype.names, item))
    return item.decode(encoding) if isinstance(item, bytes) else item


def _read_values(data: bytes, offset: int, dtype, count) -> list:
    """Reads in multiple values from a buffer starting at offset.

    Just a handy wrapper for np.frombuffer() with count >= 1.

    The read values are converted to a list of built-in datatypes using
    np.ndarray.tolist().

    Parameters
    ----------
    data
        An object that exposes the buffer interface. Here always bytes.

    offset
        Start reading the buffer from this offset (in bytes).

    dtype
        Data-type to read in.

    count
        Number of items to read. -1 means all data in the buffer.

    Returns
    -------
    Any
        The values read from the buffer as specified by the arguments.

    """
    values = np.frombuffer(data, offset=offset, dtype=dtype, count=count)
    if values.dtype.names:
        return [dict(zip(value.dtype.names, value.item())) for value in values]
    # The ndarray.tolist() method converts numpy scalars to built-in
    # scalars, hence not just list(values).
    return values.tolist()


def _process_settings(data: bytes) -> tuple[dict, list]:
    """Processes the contents of settings modules.

    Parameters
    ----------
    data
        The data to parse through.

    Returns
    -------
    dict
        The parsed settings.

    """
    settings = {}
    # First parse the settings right at the top of the data block.
    technique, params_dtype = technique_params_dtypes[data[0x0000]]
    settings["technique"] = technique
    for offset, (dtype, name) in settings_dtypes.items():
        settings[name] = _read_value(data, offset, dtype)
    # Then determine the technique parameters. The parameters' offset
    # changes depending on the technique present and apparently on some
    # other factor that is unclear to me.
    params_offset = None
    for offset in (0x0572, 0x1845, 0x1846):
        logging.debug(f"Trying to find the technique parameters at {offset:x}.")
        n_params = _read_value(data, offset + 0x0002, "<u2")
        if isinstance(params_dtype, list):
            # The params_dtype has multiple possible lengths if a list.
            for dtype in params_dtype:
                if len(dtype) == n_params:
                    params_dtype, params_offset = dtype, offset
        elif len(params_dtype) == n_params:
            params_offset = offset
            logging.debug(f"Determined {n_params} parameters at 0x{offset:x}.")
            break
    if params_offset is None:
        raise NotImplementedError("Unknown parameter offset or technique dtype.")
    logging.debug(f"Reading number of parameter sequences at 0x{params_offset:x}.")
    ns = _read_value(data, params_offset, "<u2")
    logging.debug(f"Reading {ns} parameter sequences of {n_params} parameters.")
    rawparams = _read_values(data, params_offset + 0x0004, params_dtype, ns)
    params = []
    for pardict in rawparams:
        for k, v in pardict.items():
            # MPR quirk: I_range off by one
            if k == "I_range":
                v += 1
            pardict[k] = param_from_key(k, v, to_str=True)
            # Handle NaNs and +/-Inf in params here
            if np.isnan(v) or np.isinf(v):
                pardict[k] = str(v)
        params.append(pardict)
    return settings, params


def _parse_columns(column_ids: list[int]) -> tuple[list, list, list, dict]:
    """Puts together column info from a list of data column IDs.

    Note
    ----
    The binary layout of the data in the `.mpr` file is described by a
    sequence of column IDs. Some column IDs relate to (flags) which are
    all packed into a single byte.

    Parameters
    ----------
    column_ids
        A list of column IDs.

    Returns
    -------
    tuple[list, list, list, dict]
        The column names, dtypes, units and a dictionary of flag names
        and bitmasks.

    """
    names = []
    dtypes = []
    units = []
    flags = {}
    for id in column_ids:
        if id in flag_columns:
            bitmask, name = flag_columns[id]
            flags[name] = bitmask
            # Flags column only needs to be added once.
            if "flags" not in names:
                names.append("flags")
                dtypes.append("|u1")
                units.append(None)
        elif id in data_columns:
            dtype, name, unit = data_columns[id]
            names.append(name)
            dtypes.append(dtype)
            units.append(unit)
        else:
            raise NotImplementedError(f"Unknown column ID: {id}")
    return names, dtypes, units, flags


def _process_data(
    data: bytes, version: int, Eranges: list[float], Iranges: list[float]
) -> list[dict]:
    """Processes the contents of data modules.

    Parameters
    ----------
    data
        The data to parse through.

    version
        Module version from the data module header.

    Returns
    -------
    list[dict]
        Processed data ([{column -> value}, ..., {column -> value}]). If
        the column unit is set to None, the value is an int. Otherwise,
        the value is a dict with value ("n"), sigma ("s"), and unit
        ("u").

    """
    n_datapoints = _read_value(data, 0x0000, "<u4")
    n_columns = _read_value(data, 0x0004, "|u1")
    column_ids = _read_values(data, 0x0005, "<u2", n_columns)
    # Length of each datapoint depends on number and IDs of columns.
    names, dtypes, units, flags = _parse_columns(column_ids)
    data_dtype = np.dtype(list(zip(names, dtypes)))
    # Depending on module version, datapoints start at 0x0195 or 0x0196.
    if version == 2:
        offset = 0x0195
    elif version == 3:
        offset = 0x0196
    else:
        raise NotImplementedError(f"Unknown data module version: {version}")
    datapoints = _read_values(data, offset, data_dtype, n_datapoints)
    for datapoint in datapoints:
        # Lets split this into two loops: get the indices first, then the data
        for (name, value), unit in list(zip(datapoint.items(), units)):
            if unit is None:
                intv = int(value)
                if name == "I Range":
                    datapoint[name] = param_from_key("I_range", intv)
                    Irange = param_from_key("I_range", intv, to_str=False)
                else:
                    datapoint[name] = intv
        if "Ns" in datapoint:
            Erange = Eranges[datapoint["Ns"]]
        else:
            logging.info(
                "eclab.mpr: 'Ns' is not in data table, "
                "using the first E range specified in params."
            )
            Erange = Eranges[0]
        if "I Range" not in datapoint:
            logging.info(
                "eclab.mpr: 'I Range' is not in data table, "
                "using the I range specified in params."
            )
            if "Ns" in datapoint:
                Irstr = Iranges[datapoint["Ns"]]
            else:
                Irstr = Iranges[0]
            Irange = param_from_key("I_range", Irstr, to_str=False)
        for (name, value), unit in list(zip(datapoint.items(), units)):
            if unit is None:
                continue
            s = get_resolution(name, value, Erange, Irange)
            datapoint[name] = {"n": value, "s": s, "u": unit}
        if flags:
            logging.debug("Extracting flag values.")
            flag_bits = datapoint.pop("flags")
            for name, bitmask in flags.items():
                # Two's complement hack to find the position of the
                # rightmost set bit.
                shift = (bitmask & -bitmask).bit_length() - 1
                # Rightshift flag by that amount.
                datapoint[name] = (flag_bits & bitmask) >> shift
    return datapoints


def _process_log(data: bytes) -> dict:
    """Processes the contents of log modules.

    Parameters
    ----------
    data
        The data to parse through.

    Returns
    -------
    dict
        The parsed log.

    """
    log = {}
    for offset, (dtype, name) in log_dtypes.items():
        log[name] = _read_value(data, offset, dtype)
    return log


def _process_loop(data: bytes) -> dict:
    """Processes the contents of loop modules.

    Parameters
    ----------
    data
        The data to parse through.

    Returns
    -------
    dict
        The parsed loops.

    """
    n_indexes = _read_value(data, 0x0000, "<u4")
    indexes = _read_values(data, 0x0004, "<u4", n_indexes)
    return {"n_indexes": n_indexes, "indexes": indexes}


def _process_modules(contents: bytes) -> tuple[dict, list, list, dict, dict]:
    """Handles the processing of all modules.

    Parameters
    ----------
    contents
        The contents of an .mpr file, minus the file magic.

    Returns
    -------
    tuple[dict, list, dict, dict]
        The processed settings, data, log, and loop modules. If they are
        not present in the provided modules, returns None instead.

    """
    modules = contents.split(b"MODULE")[1:]
    settings = data = log = loop = None
    for module in modules:
        header = _read_value(module, 0x0000, module_header_dtype)
        name = header["short_name"].strip()
        logging.debug(f"Read '{name}' module.")
        module_data = module[module_header_dtype.itemsize :]
        if name == "VMP Set":
            settings, params = _process_settings(module_data)
            Eranges = []
            Iranges = []
            for el in params:
                Eranges.append(el["E_range_max"] - el["E_range_min"])
                Iranges.append(el.get("I_range", "Auto"))
        elif name == "VMP data":
            data = _process_data(module_data, header["version"], Eranges, Iranges)
        elif name == "VMP LOG":
            log = _process_log(module_data)
        elif name == "VMP loop":
            loop = _process_loop(module_data)
        else:
            raise NotImplementedError(f"Unknown module: {name}.")
    return settings, params, data, log, loop


def process(
    fn: str, encoding: str = "windows-1252", timezone: str = "localtime"
) -> tuple[list, dict, bool]:
    """Processes EC-Lab raw data binary files.

    Parameters
    ----------
    fn
        The file containing the data to parse.

    encoding
        Encoding of ``fn``, by default "windows-1252".

    timezone
        A string description of the timezone. Default is "localtime".

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and the full date tag. For mpr files,
        the full date is always specified.

    """
    file_magic = b"BIO-LOGIC MODULAR FILE\x1a                         \x00\x00\x00\x00"
    with open(fn, "rb") as mpr_file:
        assert mpr_file.read(len(file_magic)) == file_magic, "invalid file magic"
        mpr = mpr_file.read()
    settings, params, data, log, loop = _process_modules(mpr)
    assert settings is not None, "no settings module"
    assert data is not None, "no data module"
    assert log is not None, "no log module"
    # Arrange all the data into the correct format.
    # TODO: Metadata could be handled in a nicer way.
    metadata = {"settings": settings, "params": params, "log": log}
    start_time = ole_to_uts(log["ole_timestamp"], timezone=timezone)
    timesteps = []
    # If the technique is an impedance spectroscopy, split it into
    # traces at different cycle numbers and put each trace into its own timestep
    if settings["technique"] in {"PEIS", "GEIS"}:
        # Grouping by cycle.
        cycles = defaultdict(list)
        for d in data:
            cycles[d["cycle number"]].append(d)
        # Casting cycles into traces.
        for ti, td in cycles.items():
            trace = {col: [d[col] for d in td] for col in td[0].keys()}
            for key, val in trace.items():
                if not isinstance(val[0], dict):
                    continue
                trace[key] = {k: [i[k] for i in val] for k in val[0]}
                # Reducing unit list to just a string.
                trace[key]["u"] = set(trace[key]["u"]).pop()
            uts = start_time + trace["time"]["n"][0]
            trace["time"]["n"] = [i - trace["time"]["n"][0] for i in trace["time"]["n"]]
            timesteps.append(
                {
                    "uts": uts,
                    "fn": fn,
                    "raw": {"traces": {settings["technique"]: trace}},
                }
            )
        return timesteps, metadata, True
    # All other techniques have multiple timesteps.
    for d in data:
        uts = start_time + d["time"]["n"]
        timesteps.append({"fn": fn, "uts": uts, "raw": d})
    return timesteps, metadata, True



"""
File parser for EZ-Chrom ASCII export files (dat.asc).

This file format includes one timestep with multiple traces in each ASCII file. It 
contains a header section, and a sequence of Y datapoints for each detector. The X
axis is uniform between traces, and its units have to be deduced from the header.

Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      sampleid: !!str
      username: !!str
      version:  !!str
      valve:    None
      datafile: !!str

.. codeauthor:: Peter Kraus <peter.kraus@empa.ch>
"""
import numpy as np
import logging
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr

from yadg.dgutils.dateutils import str_to_uts


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict, dict]:
    """
    EZ-Chrome ASCII export file parser.

    One chromatogram per file with multiple traces. A header section is followed by
    y-values for each trace. x-values have to be deduced using number of points,
    frequency, and x-multiplier. Method name is available, but detector names are not.
    They are assigned their numerical index in the file.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    ([chrom], metadata): tuple[list, dict]
        Standard timesteps & metadata tuple.
    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()
    metadata = {"filetype": "ezchrom.datasc", "params": {"valve": None}}
    chrom = {"fn": str(fn), "traces": {}}

    for line in lines:
        for key in ["Version", "Method", "User Name"]:
            if line.startswith(key):
                k = key.lower().replace(" ", "")
                metadata["params"][k] = line.split(f"{key}:")[1].strip()
        for key in ["Sample ID", "Data File"]:
            if line.startswith(key):
                k = key.lower().replace(" ", "")
                chrom[k] = line.split(f"{key}:")[1].strip()
        if line.startswith("Acquisition Date and Time:"):
            chrom["uts"] = str_to_uts(
                line.split("Time:")[1].strip(),
                format="%m/%d/%Y %I:%M:%S %p",
                timezone=timezone,
            )
        if line.startswith("Sampling Rate:"):
            assert (
                "Hz" in line
            ), f"datasc: Incorrect units for rate in file {fn}: {line}"
            parts = line.split("\t")
            samplerates = [float(each.strip()) for each in parts[1:-1]]
        if line.startswith("Total Data Points:"):
            assert (
                "Pts." in line
            ), f"datasc: Incorrect units for number of points in file {fn}: {line}"
            parts = line.split("\t")
            npoints = [int(each.strip()) for each in parts[1:-1]]
        if line.startswith("X Axis Title:"):
            parts = line.split("\t")
            xunits = [each.strip() for each in parts[1:]]
        if line.startswith("Y Axis Title:"):
            parts = line.split("\t")
            yunits = [each.strip() for each in parts[1:]]
            if "25 V" in yunits:
                yunits = [i.replace("25 V", "V") for i in yunits]
                logging.warning(
                    "ezchromasc: Implicit conversion of y-axis unit from '25 V' to 'V'."
                )
        if line.startswith("X Axis Multiplier:"):
            parts = line.split("\t")
            xmuls = [float(each.strip()) for each in parts[1:]]
        if line.startswith("Y Axis Multiplier:"):
            parts = line.split("\t")
            ymuls = [float(each.strip()) for each in parts[1:]]
        if ":" not in line:
            si = lines.index(line)
            break
    assert (
        len(samplerates)
        == len(npoints)
        == len(xunits)
        == len(yunits)
        == len(xmuls)
        == len(ymuls)
    ), f"datasc: Inconsistent number of traces in {fn}."

    for ti in range(len(samplerates)):
        assert (
            xunits[ti] == "Minutes"
        ), f"datasc: X units label of trace {ti} in {fn} was not understood."
        dt = 60
        xmul = xmuls[ti] * dt / samplerates[ti]
        ymul = ymuls[ti]
        xsn = np.arange(npoints[ti]) * xmul
        xss = np.ones(npoints[ti]) * xmul
        xs = [xsn, xss]
        ytup = [tuple_fromstr(l) for l in lines[si : si + npoints[ti]]]
        ysn, yss = [np.array(p) * ymul for p in zip(*ytup)]
        ys = [ysn, yss]
        chrom["traces"][f"{ti}"] = {"id": ti, "data": [xs, ys]}
        chrom["traces"][f"{ti}"]["t"] = {
            "n": xsn.tolist(),
            "s": xss.tolist(),
            "u": "s",
        }
        chrom["traces"][f"{ti}"]["y"] = {
            "n": ysn.tolist(),
            "s": yss.tolist(),
            "u": yunits[ti],
        }

        si += npoints[ti]
    return [chrom], metadata



import setuptools
import os

with open("VERSION", "r") as infile:
    version = infile.read().strip()

with open("README.md", "r", encoding="utf-8") as infile:
    readme = infile.read()

packagedir = "src"

setuptools.setup(
    name="yadg",
    version=version,
    author="Peter Kraus",
    author_email="peter@tondon.de",
    description="yet another datagram",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/dgbowl/yadg",
    project_urls={
        "Bug Tracker": "https://github.com/dgbowl/yadg/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": packagedir},
    packages=setuptools.find_packages(where=packagedir),
    python_requires=">=3.8",
    install_requires=[
        "numpy",
        "scipy",
        "pint",
        "uncertainties",
        "striprtf",
        "pytest",
        "tzlocal",
        "python-dateutil",
    ],
    entry_points={"console_scripts": ["yadg=yadg:run_with_arguments"]},
)



"""
.. _parsers_masstrace_formats:

List of supported file formats:

 - Pfeiffer Quadstar 32-bit scan analog data (sac) :mod:`yadg.parsers.masstrace.quadstarsac`
"""
from .main import process, version



datagram_version = "4.0.rc1"

datagram_step = {
    "type": dict,
    "all": {
        "metadata": {
            "type": dict,
            "all": {
                "tag": {"type": str},
                "parser": {
                    "type": dict,
                    "each": {
                        "type": dict,
                        "all": {"version": {"type": str}},
                        "allow": True,
                    },
                },
            },
            "allow": True,
        },
        "data": {
            "type": list,
            "each": {
                "type": dict,
                "all": {"uts": {"type": float}},
                "any": {
                    "fn": {"type": str},
                    "raw": {"type": dict},
                    "derived": {"type": dict},
                },
            },
        },
    },
}

datagram = {
    "type": dict,
    "all": {
        "metadata": {
            "type": dict,
            "all": {
                "provenance": {"type": dict},
                "date": {"type": str},
                "input_schema": {"type": dict},
                "datagram_version": {"type": str},
            },
        },
        "steps": {"type": list, "each": datagram_step},
    },
}



import numpy as np
from io import BufferedIOBase
from typing import Union


def read_from_file(
    f: BufferedIOBase, offset: int, dtype: str, count: int = 1
) -> Union[str, np.ndarray]:
    f.seek(offset)
    if dtype == "utf-8":
        len = int.from_bytes(f.read(1), byteorder="big")
        chars = f.read(len)
        return chars.decode("utf-8")
    elif dtype == "utf-16":
        len = int.from_bytes(f.read(1), byteorder="big") * 2
        chars = f.read(len)
        return chars.decode("utf-16")
    elif count == 1:
        return np.fromfile(f, offset=0, dtype=dtype, count=1)[0]
    else:
        return np.fromfile(f, offset=0, dtype=dtype, count=count)


def read_from_buffer(
    buf: bytes, offset: int, dtype: str, count: int = 1
) -> Union[str, np.ndarray]:
    if dtype == "utf-8":
        len = int.from_bytes(buf[offset], byteorder="big")
        chars = buf[offset + 1 : offset + 1 + len]
        return chars.decode("utf-8")
    elif dtype == "utf-16":
        len = int.from_bytes(buf[offset], byteorder="big") * 2
        chars = buf[offset + 1 : offset + 1 + len]
        return chars.decode("utf-16")
    elif count == 1:
        return np.frombuffer(buf, offset=offset, dtype=dtype, count=1)[0]
    else:
        return np.frombuffer(buf, offset=offset, dtype=dtype, count=count)


def read_value(
    object: Union[BufferedIOBase, bytes], offset: int, dtype: str, count: int = 1
) -> Union[str, np.ndarray, int, float]:
    if isinstance(object, bytes):
        return read_from_buffer(object, offset, dtype, count)
    else:
        return read_from_file(object, offset, dtype, count)



from yadg.parsers.xpstrace import phispe


version = "4.0.0"


def process(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "UTC",
    tracetype: str = "phi.spe",
) -> tuple[list, dict, bool]:
    """Unified x-ray photoelectron spectroscopy parser.

    This parser processes XPS scans in signal(energy) format.

    Parameters
    ----------
    fn
        The file containing the trace(s) to parse.

    encoding
        Encoding of ``fn``, by default "utf-8".

    timezone
        A string description of the timezone. Default is "UTC".

    tracetype
        Determines the output file format. Currently supported formats
        can be found :ref:`here<parsers_xpstrace_formats>`.

        The default is ``"phi.spe"``.

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and the full date tag.
        Multipak .spe files seemingly have no timestamp.

    """
    if tracetype == "phi.spe":
        return phispe.process(fn, encoding, timezone)



"""
File parser for DryCal log files, including converted documents (rtf) and tabulated
exports (txt, csv).

The DryCal files only contain the timestamps of the datapoints, not the date. Therefore,
the date has to be supplied either using the ``date`` argument in parameters, or is 
parsed from the prefix of the filename.

.. codeauthor:: Peter Kraus <peter.kraus@empa.ch>
"""

from yadg.parsers.basiccsv import process_row
from striprtf.striprtf import rtf_to_text
import yadg.dgutils


def rtf(
    fn: str,
    encoding: str = "utf-8",
    timezone: str = "UTC",
    calib: dict = {},
) -> tuple[list, dict]:
    """
    RTF version of the drycal parser.

    This is intended to parse legacy drycal DOC files, which have been converted to RTF
    using other means.

    Parameters
    ----------
    fn
        Filename to parse.

    encoding
        Encoding to use for parsing ``fn``.

    calib
        A calibration spec.

    Returns
    -------
    (timesteps, metadata, None): tuple[list, dict, None]
        A standard data - metadata - common data output tuple.
    """
    with open(fn, "r", encoding=encoding) as infile:
        rtf = infile.read()
    lines = rtf_to_text(rtf).split("\n")
    for li in range(len(lines)):
        if lines[li].startswith("Sample"):
            si = li
        elif lines[li].startswith("1|"):
            di = li
            break
    # Metadata processing for rtf files is in columns, not rows.
    ml = []
    metadata = dict()
    for line in lines[:si]:
        if line.strip() != "":
            items = [i.strip() for i in line.split("|")]
            if len(items) > 1:
                ml.append(items)
    assert len(ml) == 2 and len(ml[0]) == len(ml[1])
    for i in range(len(ml[0])):
        if ml[0][i] != "":
            metadata[ml[0][i]] = ml[1][i]

    # Process data table
    dl = []
    dl.append(" ".join(lines[si:di]))
    for line in lines[di:]:
        if line.strip() != "":
            dl.append(line)
    headers, units, data = drycal_table(dl, sep="|")
    datecolumns, datefunc, _ = yadg.dgutils.infer_timestamp_from(
        spec={"time": {"index": 4, "format": "%I:%M:%S %p"}}, timezone=timezone
    )

    # Correct each ts by provided date
    timesteps = []
    for r in data:
        ts = process_row(headers[1:], r[1:], units, datefunc, datecolumns, calib=calib)
        ts["fn"] = fn
        timesteps.append(ts)

    return timesteps, metadata


def sep(
    fn: str,
    sep: str,
    encoding: str = "utf-8",
    timezone: str = "UTC",
    calib: dict = {},
) -> tuple[list, dict]:
    """
    Generic drycal parser, using ``sep`` as separator string.

    This is intended to parse other export formats from DryCal, such as txt and csv files.

    Parameters
    ----------
    fn
        Filename to parse.

    date
        A unix timestamp float corresponding to the day (or other offset) to be added to
        each line in the measurement table.

    sep
        The separator character used to split lines in ``fn``.

    encoding
        Encoding to use for parsing ``fn``.

    calib
        A calibration spec.

    Returns
    -------
    (timesteps, metadata, None): tuple[list, dict, None]
        A standard data - metadata - common data output tuple.
    """
    with open(fn, "r", encoding=encoding) as infile:
        lines = infile.readlines()
    for li in range(len(lines)):
        if lines[li].startswith("Sample"):
            si = li
        elif lines[li].startswith(f"1{sep}"):
            di = li
            break
    # Metadata processing for csv files is standard.
    metadata = dict()
    for line in lines[:si]:
        if line.strip() != "":
            items = [i.strip() for i in line.split(sep)]
            if len(items) == 2:
                metadata[items[0]] = items[1]

    # Process data table
    dl = list()
    dl.append(" ".join(lines[si:di]))
    for line in lines[di:]:
        if line.strip() != "":
            dl.append(line)
    headers, units, data = drycal_table(dl, sep=sep)

    if "AM" in data[0][-1].upper() or "PM" in data[0][-1].upper():
        fmt = "%I:%M:%S %p"
    else:
        fmt = "%H:%M:%S"
    datecolumns, datefunc, _ = yadg.dgutils.infer_timestamp_from(
        spec={"time": {"index": 4, "format": fmt}}, timezone=timezone
    )

    # Correct each ts by provided date
    timesteps = list()
    for r in data:
        ts = process_row(headers[1:], r[1:], units, datefunc, datecolumns, calib=calib)
        ts["fn"] = str(fn)
        timesteps.append(ts)

    return timesteps, metadata


def drycal_table(lines: list, sep: str = ",") -> tuple[list, dict, list]:
    """
    DryCal table-processing function.

    Given a table with headers and units in the first line, and data in the following
    lines, this function returns the headers, units, and data extracted from the table.
    The returned values are always of :class:`(str)` type, any post-processing is done
    in the calling routine.

    Parameters
    ----------
    lines
        A list containing the lines to be parsed

    sep
        The separator string used to split each line into individual items

    Returns
    -------
    (headers, units, data): tuple[list, dict, list]
        A tuple of a list of the stripped headers, dictionary of header-unit key-value
        pairs, and a list of lists containing the rows of the table.
    """
    items = [i.strip() for i in lines[0].split(sep)]
    headers = []
    units = {}
    data = []
    for item in items:
        for rs in [". ", " "]:
            parts = item.split(rs)
            if len(parts) == 2:
                break
        headers.append(parts[0])
        if len(parts) == 2:
            units[parts[0]] = parts[1]
        else:
            units[parts[0]] = " "
    if items[-1] == "":
        trim = True
        headers = headers[:-1]
    for line in lines[1:]:
        cols = line.split(sep)
        assert len(cols) == len(items)
        if trim:
            data.append(cols[:-1])
        else:
            data.append(cols)

    yadg.dgutils.sanitize_units(units)
    return headers, units, data



"""
.. _parsers_electrochem_formats:

List of supported file formats:

 - EC-Lab raw data binary file and parameter settings 
   :mod:`yadg.parsers.electrochem.eclabmpr`
 - EC-Lab human-readable text export of data 
   :mod:`yadg.parsers.electrochem.eclabmpt`
"""
from .main import process, version



import numpy as np
import logging
import os
from typing import Union

import yadg.core
from yadg.dgutils import ureg


def _list_validator(l: list) -> bool:
    if len(l) == 3:
        return _float_list(l)
    else:
        return _general_list(l)


def _float_list(l: list) -> bool:
    if isinstance(l[0], float) and isinstance(l[1], float) and isinstance(l[2], str):
        logging.warning("Old [float, float, str] syntax detected.")
        return True
    else:
        return _general_list(l)


def _general_list(l: list) -> bool:
    for v in l:
        if isinstance(v, list):
            return _list_validator(v)
        elif isinstance(v, dict):
            return _dict_validator(v)
        else:
            assert isinstance(v, str) or isinstance(v, int), (
                f"List elements have to be one of [str, int, dict, list], "
                f"but entry id:{l.index(v)}:{type(v)} is neither: {l}"
            )
    return True


def _unit_validator(u: str) -> bool:
    assert u in ureg, f"unit {u} is not defined."
    return True


def _dict_validator(d: dict) -> bool:
    for k, v in d.items():
        if k in ["n", "s", "u"] and len({"n", "s", "u"}.intersection(d.keys())) == 3:
            if k in ["n", "s"] and isinstance(v, (float, list)):
                pass
            elif k == "u" and isinstance(v, str):
                assert _unit_validator(v)
            else:
                logging.error(f"validator: we shouldn't be here: {k}, {v}")
                return False
        elif isinstance(v, float):
            assert k == "uts", f"Only 'uts' can be a float entry, not '{k}'."
        elif isinstance(v, list):
            assert _list_validator(v)
        elif isinstance(v, dict):
            assert _dict_validator(v)
        else:
            assert isinstance(v, str) or isinstance(v, int), (
                f"Dict elements have to be one of [str, int, dict, list], "
                f"but '{k}':{type(v)} is neither: {v}"
            )
    return True


def validator(item: Union[list, dict, str], spec: dict) -> True:
    """
    Worker validator function.

    This function checks that ``item`` matches the specification supplied in ``spec``.
    The ``spec`` :class:`(dict)` can have the following entries:

    - ``"type"`` :class:`(type)`, a required entry, defining the type of ``item``,
    - ``"all"`` :class:`(dict)` defining a set of required keywords and their respective ``spec``,
    - ``"any"`` :class:`(dict)` defining a set of optional keywords and their respective ``spec``,
    - ``"one"`` :class:`(dict)` defining a set of mutually exclusive keywords and their respective ``spec``,
    - ``"each"`` :class:`(dict)` providing the ``spec`` for any keywords not listed in ``"all"``, ``"any"``, or ``"one"``,
    - ``"allow"`` :class:`(bool)` a switch whether to allow unspecified keys.

    To extend the existing `datagram` and `schema` specs, look into :mod:`yadg.core.spec_datagram`
    and :mod:`yadg.core.spec_schema`, respectively.

    Parameters
    ----------
    item
        The ``item`` to be validated.

    spec
        The ``spec`` with which to validate the ``item``

    Returns
    -------
    True: bool
        If the ``item`` matches the ``spec``, returns `True`. Otherwise, an `AssertionError` is raised.
    """
    assert isinstance(item, spec["type"]), (
        f"validator: item '{item}' does not match prescribed type "
        f"in spec '{spec['type']}'."
    )
    if len({"all", "one", "any", "each"}.intersection(spec)) > 0 and spec["type"] in [
        list,
        dict,
    ]:
        for k, v in spec.get("all", {}).items():
            assert k in item, (
                f"validator: required entry '{k}' was not specified "
                f"in item '{item}'."
            )
        if "one" in spec:
            initem = set(spec["one"]).intersection(item)
            assert len(initem) == 1, (
                f"validator: Exactly one of entries in {spec['one']} has to be "
                f"provided in item {item}, but {len(initem)} were provided: {initem}."
            )
        for k in item:
            s = False
            for d in ["all", "one", "any"]:
                if k in spec.get(d, []):
                    s = d
            assert s or spec.get("allow", False) or "each" in spec, (
                f"validator: Key '{k}' in item {item} " f"is not understood."
            )
            if s:
                assert validator(item[k], spec[s][k])
            elif "each" in spec:
                if spec["type"] == list:
                    assert validator(k, spec["each"])
                else:
                    assert validator(item[k], spec["each"])
    elif len({"all", "one", "any"}.intersection(spec)) > 0 and spec["type"] in [str]:
        if "all" in spec:
            assert (
                len(spec["all"]) == 1 and item in spec["all"]
            ), f"validator: Item '{item}' is not in {spec['all']}."
        if "one" in spec:
            assert (
                item in spec["one"]
            ), f"validator: Item '{item}' is not in {spec['one']}."
        if "any" in spec:
            assert (
                item in spec["any"]
            ), f"validator: Item '{item}' is not in {spec['any']}."
    return True


def validate_schema(
    schema: Union[list, tuple], strictfiles: bool = True, strictfolders: bool = True
) -> True:
    """
    Schema validator.

    Checks the overall `schema` format, checks every `step` of the `schema` for required
    entries, and checks whether required parameters for each `parser` are provided. The
    validator additionally fills in optional parameters, where necessary for a valid `schema`.

    The specification is:

    - The `schema` has to be a :class:`(dict)` with a top-level ``"metadata"``
      :class:`(dict)` and ``"steps"`` :class:`(list)` entries
    - The ``"metadata"`` entry has to specify the ``"provenance"`` of the `schema`, as
      well as the ``"schema_version"`` :class:`(str)`. Other entries may be specified.
    - Each element within the ``"steps"`` list is a `step`, of type :class:`(dict)`
    - Each `step` has to have the ``"parser"`` and ``"import"`` entries:

      - The ``"parser"`` is a :class:`(str)` entry that has to contain the name of the requested
        parser module. This entry is processed in the :func:`yadg.core.main._infer_datagram_handler`
        function in the core module.
      - The ``"import"`` is a :class:`(dict)` entry has to contain:

        - Exactly **one** entry out of ``"files"`` or ``"folders"``. This entry must be a
          :class:`(list)` even if only one element is provided.
        - Any combination of ``"prefix"``, ``"suffix"``, ``"contains"``, and ``"exclude"``
          entries. They must be of type :class:`(str)`. These entries specify the matching
          of files within folders accordingly.

    - The only other allowed entries are:

      - ``"tag"`` :class:`(str)`: for defining a tag for each `step`; by default
        assigned the numerical index of the `step` within the `schema`.
      - ``"export"`` :class:`(str)`: for exporting a single `step`; whether the processed
        `step` should be exported as a ``json`` file. This file is kept available for other
        `steps`, but will be removed at the end of `schema` processing.
      - ``"parameters"`` :class:`(dict)`: for specifying other parameters for each of the parsers.

    - no other entries are permitted

    Parameters
    ----------
    schema
        The schema to be validated.

    strictfiles
        When `False`, any files specified using the ``"files"`` option will not be
        checked for existence. Note that folders (specified via ``"folders"``) are always checked.

    Returns
    -------
    True: bool
        When the `schema` is valid and passes all assertions, `True` is returned. Otherwise,
        an `AssertionError` is raised.
    """
    # schema has to meet the spec
    assert validator(schema, yadg.core.spec_schema.schema)
    # log default timezone
    if "timezone" not in schema["metadata"]:
        logging.warning(f"schema_validator: Timezone not specified. Using 'localtime'.")
    for step in schema["steps"]:
        si = schema["steps"].index(step)
        # import files or folders must exist
        if "files" in step["import"] and strictfiles:
            for fn in step["import"]["files"]:
                assert os.path.exists(fn) and os.path.isfile(
                    fn
                ), f"schema_validator: File path {fn} provided in step {si} is not a valid file."
        if "folders" in step["import"] and strictfolders:
            for fn in step["import"]["folders"]:
                assert os.path.exists(fn) and os.path.isdir(
                    fn
                ), f"schema_validator: Folder path {fn} provided in step {si} is not a valid folder."
        # log default tag
        if "tag" not in step:
            logging.info(
                f"schema_validator: Tag not present in step {si}. Using '{si:02d}'"
            )
        # log default encoding
        if "encoding" not in step["import"]:
            logging.info(
                f"schema_validator: Encoding not present in step {si}. Using 'utf-8'"
            )
    return True


def validate_datagram(datagram: dict) -> True:
    """
    Datagram validator.

    Checks the overall `datagram` format against the `datagram` spec, and ensures that
    each floating-point value is accompanied by standard deviation and unit.

    The current `datagram` specification is:

    -  The `datagram` must be a :class:`(dict)` with two entries:

       - ``"metadata"`` :class:`(dict)`: A top-level entry containing metadata.
       - ``"steps"`` :class:`(list[dict])`: List corresponding to a sequence of `steps`.

    - The ``"metadata"`` entry has to contain information about the ``"provenance"`` of
      the `datagram`, the creation date using ISO8601 format in ``"date"`` :class:`(str)`
      entry, a full copy of the input `schema` in the ``"input_schema"`` entry, and
      version information in ``"datagram_version"`` :class:`(str)`.

    - Each element in the ``"steps"`` corresponds to a single `step` from the `schema`.
    - Each `step` in ``"steps"`` has to contain a ``"metadata"`` :class:`(dict)` entry,
      and a ``"data"`` :class:`(list)`; an optional ``"common"`` :class:`(dict)` data
      block can be provided.
    - Each timestep in the ``"data"`` list has to specify a timestamp using the
      Unix Timestamp format in ``"uts"`` :class:`(float)` entry; the original filename
      in ``"fn"`` :class:`(str)` entry. The raw data present in this original filename
      is stored as sub-entries within the ``"raw"`` :class:`(dict)` entry. Any derived
      data, such as that obtained via calibration, integration, or fitting, has to be
      stored in the ``"derived"`` :class:`(dict)`.

    .. note::
        A floating-point entry should always have its standard deviation specified.
        Internal processing of this data is always carried out using the :class:`(ufloat)`
        type, which ought to be exported as a ``{"n": value, "s": std_dev, "u": "-"}`` keypair.

    .. note::
        Most numerical data should have associated units. The validator expects all
        floating-point entries to be in a ``{"n": value, "s": std_dev}`` format for
        properties with an arbitrary unit, and ``{"n": value, "s": std_dev, "u": unit}``
        for properties with a defined unit.

    Parameters
    ----------
    datagram
        The `datagram` to be validated.

    Returns
    -------
    True: bool
        If the `datagram` passes all assertions, returns `True`. Else, an `AssertionError` is raised.
    """
    # datagram has to meet the spec
    assert validator(datagram, yadg.core.spec_datagram.datagram)
    # validate each step in the datagram
    for step in datagram["steps"]:
        assert all(
            ["fn" in ts for ts in step["data"]]
        ), "The 'fn' entry has to be provided in each timestep."
        for ts in step["data"]:
            assert _dict_validator(ts)
    return True



"""
.. _parsers_chromtrace_formats:

List of supported file formats:

 - EZ-Chrom ASCII export (dat.asc) :mod:`yadg.parsers.chromtrace.ezchromasc`
 - Agilent Chemstation Chromtab (csv) :mod:`yadg.parsers.chromtrace.agilentcsv`
 - Agilent OpenLab binary signal (ch) :mod:`yadg.parsers.chromtrace.agilentch` 
 - Agilent OpenLab data archive (dx) :mod:`yadg.parsers.chromtrace.agilentdx` 
 - Fusion JSON format (json) :mod:`yadg.parsers.chromtrace.fusionjson` 
 - Fusion zip archive (zip) :mod:`yadg.parsers.chromtrace.fusionjson`
"""
from .main import process, version



"""Processing of BioLogic's EC-Lab ASCII export files.

File Structure of `.mpt` Files
``````````````````````````````

These human-readable files are sectioned into headerlines and datalines.
The header part at is made up of information that can be found in the
settings, log and loop modules of the binary `.mpr` file.

If no header is present, the timestamps will instead be calculated from
the file's ctime.


Structure of Parsed Data
````````````````````````

*EIS Techniques (PEIS/GEIS)*

.. code-block:: yaml

    - fn   !!str
    - uts  !!float
    - raw:
        traces:
          "{{ technique name }}":
            "{{ col1 }}":
              [!!int, ...]
            "{{ col2 }}":
              {n: [!!float, ...], s: [!!float, ...], u: !!str}

*All Other Techniques*

.. code-block:: yaml

    - fn   !!str
    - uts  !!float
    - raw:
        "{{ col1 }}":  !!int
        "{{ col2 }}":
          {n: !!float, s: !!float, u: !!str}

.. codeauthor:: Nicolas Vetsch <vetschnicolas@gmail.com>
"""
import math
import os
import re
import logging
from collections import defaultdict
from yadg.dgutils.dateutils import str_to_uts

from yadg.parsers.electrochem.eclabtechniques import (
    get_resolution,
    technique_params,
    param_from_key,
)

# Maps EC-Lab's "canonical" column names to yadg name and unit.
column_units = {
    '"Ri"/Ohm': ("'Ri'", "Ω"),
    "-Im(Z)/Ohm": ("-Im(Z)", "Ω"),
    "-Im(Zce)/Ohm": ("-Im(Zce)", "Ω"),
    "-Im(Zwe-ce)/Ohm": ("-Im(Zwe-ce)", "Ω"),
    "(Q-Qo)/C": ("(Q-Qo)", "C"),
    "(Q-Qo)/mA.h": ("(Q-Qo)", "mA·h"),
    "<Ece>/V": ("<Ece>", "V"),
    "<Ewe>/V": ("<Ewe>", "V"),
    "<I>/mA": ("<I>", "mA"),
    "|Ece|/V": ("|Ece|", "V"),
    "|Energy|/W.h": ("|Energy|", "W·h"),
    "|Ewe|/V": ("|Ewe|", "V"),
    "|I|/A": ("|I|", "A"),
    "|Y|/Ohm-1": ("|Y|", "S"),
    "|Z|/Ohm": ("|Z|", "Ω"),
    "|Zce|/Ohm": ("|Zce|", "Ω"),
    "|Zwe-ce|/Ohm": ("|Zwe-ce|", "Ω"),
    "Analog IN 1/V": ("Analog IN 1", "V"),
    "Analog IN 2/V": ("Analog IN 2", "V"),
    "Capacitance charge/µF": ("Capacitance charge", "µF"),
    "Capacitance discharge/µF": ("Capacitance discharge", "µF"),
    "Capacity/mA.h": ("Capacity", "mA·h"),
    "charge time/s": ("charge time", "s"),
    "Conductivity/S.cm-1": ("Conductivity", "S/cm"),
    "control changes": ("control changes", None),
    "control/mA": ("control_I", "mA"),
    "control/V": ("control_V", "V"),
    "control/V/mA": ("control_V/I", "V/mA"),
    "counter inc.": ("counter inc.", None),
    "Cp-2/µF-2": ("Cp⁻²", "µF⁻²"),
    "Cp/µF": ("Cp", "µF"),
    "Cs-2/µF-2": ("Cs⁻²", "µF⁻²"),
    "Cs/µF": ("Cs", "µF"),
    "cycle number": ("cycle number", None),
    "cycle time/s": ("cycle time", "s"),
    "d(Q-Qo)/dE/mA.h/V": ("d(Q-Qo)/dE", "mA·h/V"),
    "dI/dt/mA/s": ("dI/dt", "mA/s"),
    "discharge time/s": ("discharge time", "s"),
    "dQ/C": ("dQ", "C"),
    "dq/mA.h": ("dq", "mA·h"),
    "dQ/mA.h": ("dQ", "mA·h"),
    "Ece/V": ("Ece", "V"),
    "Ecell/V": ("Ecell", "V"),
    "Efficiency/%": ("Efficiency", "%"),
    "Energy charge/W.h": ("Energy charge", "W·h"),
    "Energy discharge/W.h": ("Energy discharge", "W·h"),
    "Energy/W.h": ("Energy", "W·h"),
    "error": ("error", None),
    "Ewe-Ece/V": ("Ewe-Ece", "V"),
    "Ewe/V": ("Ewe", "V"),
    "freq/Hz": ("freq", "Hz"),
    "half cycle": ("half cycle", None),
    "I Range": ("I Range", None),
    "I/mA": ("I", "mA"),
    "Im(Y)/Ohm-1": ("Im(Y)", "S"),
    "mode": ("mode", None),
    "Ns changes": ("Ns changes", None),
    "Ns": ("Ns", None),
    "NSD Ewe/%": ("NSD Ewe", "%"),
    "NSD I/%": ("NSD I", "%"),
    "NSR Ewe/%": ("NSR Ewe", "%"),
    "NSR I/%": ("NSR I", "%"),
    "ox/red": ("ox/red", None),
    "P/W": ("P", "W"),
    "Phase(Y)/deg": ("Phase(Y)", "deg"),
    "Phase(Z)/deg": ("Phase(Z)", "deg"),
    "Phase(Zce)/deg": ("Phase(Zce)", "deg"),
    "Phase(Zwe-ce)/deg": ("Phase(Zwe-ce)", "deg"),
    "Q charge/discharge/mA.h": ("Q charge/discharge", "mA·h"),
    "Q charge/mA.h": ("Q charge", "mA·h"),
    "Q charge/mA.h/g": ("Q charge", "mA·h/g"),
    "Q discharge/mA.h": ("Q discharge", "mA·h"),
    "Q discharge/mA.h/g": ("Q discharge", "mA·h/g"),
    "R/Ohm": ("R", "Ω"),
    "Rcmp/Ohm": ("Rcmp", "Ω"),
    "Re(Y)/Ohm-1": ("Re(Y)", "S"),
    "Re(Y)/Ohm-1": ("Re(Y)", "S"),
    "Re(Z)/Ohm": ("Re(Z)", "Ω"),
    "Re(Z)/Ohm": ("Re(Z)", "Ω"),
    "Re(Zce)/Ohm": ("Re(Zce)", "Ω"),
    "Re(Zce)/Ohm": ("Re(Zce)", "Ω"),
    "Re(Zwe-ce)/Ohm": ("Re(Zwe-ce)", "Ω"),
    "Re(Zwe-ce)/Ohm": ("Re(Zwe-ce)", "Ω"),
    "step time/s": ("step time", "s"),
    "THD Ewe/%": ("THD Ewe", "%"),
    "THD I/%": ("THD I", "%"),
    "time/s": ("time", "s"),
    "x": ("x", " "),
    "z cycle": ("z cycle", None),
}


def _process_header(lines: list[str], timezone: str) -> tuple[dict, list, dict]:
    """Processes the header lines.

    Parameters
    ----------
    lines
        The header lines, starting at line 3 (which is an empty line),
        right after the `"Nb header lines : "` line.

    Returns
    -------
    tuple[dict, dict]
        A dictionary containing the settings (and the technique
        parameters) and a dictionary containing the loop indexes.

    """
    sections = "\n".join(lines).split("\n\n")
    # Can happen that no settings are present but just a loops section.
    assert not sections[1].startswith("Number of loops : "), "no settings present"
    # Again, we need the acquisition time to get timestamped data.
    assert len(sections) >= 3, "no settings present"
    technique = sections[1].strip()
    settings_lines = sections[2].split("\n")
    technique, params_keys = technique_params(technique, settings_lines)
    params = settings_lines[-len(params_keys) :]
    # The sequence param columns are always allocated 20 characters.
    n_sequences = int(len(params[0]) / 20)
    params_values = []
    for seq in range(1, n_sequences):
        values = []
        for param in params:
            try:
                val = float(param[seq * 20 : (seq + 1) * 20])
            except ValueError:
                val = param[seq * 20 : (seq + 1) * 20].strip()
            values.append(val)
        params_values.append(values)
    params = [dict(zip(params_keys, values)) for values in params_values]
    settings_lines = [line.strip() for line in settings_lines[: -len(params_keys)]]

    # Parse the acquisition timestamp.
    timestamp_re = re.compile(r"Acquisition started on : (?P<val>.+)")
    timestamp_match = timestamp_re.search("\n".join(settings_lines))
    timestamp = timestamp_match["val"]
    for format in ("%m/%d/%Y %H:%M:%S", "%m.%d.%Y %H:%M:%S", "%m/%d/%Y %H:%M:%S.%f"):
        uts = str_to_uts(timestamp, format, timezone, False)
        if uts is not None:
            break
    if uts is None:
        raise NotImplementedError(f"Time format for {timestamp} not implemented.")

    loops = None
    if len(sections) >= 4 and sections[-1].startswith("Number of loops : "):
        # The header contains a loops section.
        loops_lines = sections[-1].split("\n")
        n_loops = int(loops_lines[0].split(":")[-1])
        indexes = []
        for n in range(n_loops):
            index = loops_lines[n + 1].split("to")[0].split()[-1]
            indexes.append(int(index))
        loops = {"n_loops": n_loops, "indexes": indexes}
    settings = {
        "posix_timestamp": uts,
        "technique": technique,
        "raw": "\n".join(lines),
    }
    return settings, params, loops


def _process_data(lines: list[str], Eranges: list[float], Iranges: list[float]) -> dict:
    """Processes the data lines.

    Parameters
    ----------
    lines
        The data lines, starting right after the last header section.
        The first line is an empty line, the column names can be found
        on the second line.

    Returns
    -------
    dict
        A dictionary containing the datapoints in the format
        ([{column -> value}, ..., {column -> value}]). If the column
        unit is set to None, the value is an int. Otherwise, the value
        is a dict with value ("n"), sigma ("s"), and unit ("u").

    """
    # At this point the first two lines have already been read.
    # Remove extra column due to an extra tab in .mpt file column names.
    names = lines[1].split("\t")[:-1]
    columns, units = zip(*[column_units[n] for n in names])
    data_lines = lines[2:]
    datapoints = []
    for line in data_lines:
        values = line.split("\t")
        datapoint = {}
        for col, val, unit in list(zip(columns, values, units)):
            if unit is None:
                ival = int(float(val))
                if col == "I Range":
                    datapoint[col] = param_from_key("I_range", ival)
                    Irange = param_from_key("I_range", ival, to_str=False)
                else:
                    datapoint[col] = ival
        if "Ns" in datapoint:
            Erange = Eranges[datapoint["Ns"]]
        else:
            logging.info(
                "eclab.mpr: 'Ns' is not in data table, "
                "using the first E range specified in params."
            )
            Erange = Eranges[0]
        if "I Range" not in datapoint:
            logging.info(
                "eclab.mpr: 'I Range' is not in data table, "
                "using the I range specified in params."
            )
            if "Ns" in datapoint:
                Irstr = Iranges[datapoint["Ns"]]
            else:
                Irstr = Iranges[0]
            Irange = param_from_key("I_range", Irstr, to_str=False)
        for col, val, unit in list(zip(columns, values, units)):
            try:
                val = float(val)
            except ValueError:
                val = val.strip()
            if unit is None:
                continue
            assert isinstance(val, float), "`n` should not be string"
            s = get_resolution(col, val, Erange, Irange)
            datapoint[col] = {"n": val, "s": s, "u": unit}
        datapoints.append(datapoint)
    return datapoints


def process(
    fn: str, encoding: str = "windows-1252", timezone: str = "UTC"
) -> tuple[list, dict, bool]:
    """Processes EC-Lab human-readable text export files.

    Parameters
    ----------
    fn
        The file containing the data to parse.

    encoding
        Encoding of ``fn``, by default "windows-1252".

    timezone
        A string description of the timezone. Default is "UTC".

    Returns
    -------
    (data, metadata, fulldate) : tuple[list, dict, bool]
        Tuple containing the timesteps, metadata, and the full date tag. For mpt files,
        the full date might not be specified if header is not present.

    """
    file_magic = "EC-Lab ASCII FILE\n"
    with open(fn, "r", encoding=encoding) as mpt_file:
        assert mpt_file.read(len(file_magic)) == file_magic, "invalid file magic"
        mpt = mpt_file.read()
    lines = mpt.split("\n")
    nb_header_lines = int(lines[0].split()[-1])
    header_lines = lines[: nb_header_lines - 3]
    data_lines = lines[nb_header_lines - 3 :]
    settings, params, loops = {}, [], {}
    if nb_header_lines <= 3:
        logging.warning("eclabmpt: Header contains no settings and hence no timestamp.")
        start_time = 0.0
        fulldate = False
        Eranges = [20.0]
        Iranges = ["Auto"]
    else:
        settings, params, loops = _process_header(header_lines, timezone)
        start_time = settings.get("posix_timestamp")
        fulldate = True
        Eranges = []
        Iranges = []
        for el in params:
            Eranges.append(el["E_range_max"] - el["E_range_min"])
            Iranges.append(el.get("I_range", "Auto"))
    data = _process_data(data_lines, Eranges, Iranges)
    # Arrange all the data into the correct format.
    # TODO: Metadata could be handled in a nicer way.
    metadata = {"settings": settings, "params": params}
    timesteps = []
    # If the technique is an impedance spectroscopy, split it into
    # traces at different cycle numbers and put each trace into its own timestep
    if settings["technique"] in {"PEIS", "GEIS"}:
        # Grouping by cycle.
        cycles = defaultdict(list)
        for d in data:
            cycles[d["cycle number"]].append(d)
        # Casting cycles into traces.
        for ti, td in cycles.items():
            trace = {col: [d[col] for d in td] for col in td[0].keys()}
            for key, val in trace.items():
                if not isinstance(val[0], dict):
                    continue
                trace[key] = {k: [i[k] for i in val] for k in val[0]}
                # Reducing unit list to just a string.
                trace[key]["u"] = set(trace[key]["u"]).pop()
            uts = start_time + trace["time"]["n"][0]
            trace["time"]["n"] = [i - trace["time"]["n"][0] for i in trace["time"]["n"]]
            timesteps.append(
                {
                    "uts": uts,
                    "fn": fn,
                    "raw": {"traces": {settings["technique"]: trace}},
                }
            )
        return timesteps, metadata, fulldate
    # All other techniques have multiple timesteps.
    for d in data:
        uts = start_time + d["time"]["n"]
        timesteps.append({"fn": fn, "uts": uts, "raw": d})
    return timesteps, metadata, fulldate



"""
Quality factor fitting routines.

.. _parsers_qftrace_methods:

List of supported quality factor fitting methods:

 - Naive fitting using half-width-half-maximum around the minimum of 
   :math:`|\\Gamma(f)|` in :func:`yadg.parsers.qftrace.fit.naive`
 - Lorentzian fit to the trace of :math:`|\\Gamma(f)|` trimmed using a 
   gradient threshold criterium in :func:`yadg.parsers.qftrace.fit.lorentz`
 - Kajfez's circle fitting routine to :math:`\\Gamma(f)`, using a peak-height 
   cutoff pruning method in :func:`yadg.parsers.qftrace.fit.kajfez`

Kajfez's method is the preferred one and hence set as default, with a ``cutoff``
set to 0.4.

.. codeauthor:: Peter Kraus <peter.kraus@empa.ch>
"""
import numpy as np
from scipy.optimize import curve_fit
import uncertainties as uc


def naive(
    fvals: np.ndarray,
    fsigs: np.ndarray,
    gvals: np.ndarray,
    absgvals: np.ndarray,
    **kwargs: dict
) -> tuple[uc.ufloat, uc.ufloat]:
    """
    Naive fitting routine.

    This fit finds the central frequency :math:`f_0`, determines the full-width at
    the half-maximum :math:`\\Delta f_{HM}` by linear interpolation, and calculates
    the quality factor using :math:`Q_0 = f_0 / \\Delta f_{HM}`.

    - This fit first normalises the :math:`|\\Gamma(f)|` trace to values between 0 and 1.
    - The trace is flipped in this process, obtaining :math:`||\\Gamma(f)||`.
    - The :math:`f_0` is determined by finding the frequency corresponding to the
      minimum in :math:`|\\Gamma(f)|` (that is, the maximum of :math:`||\\Gamma(f)||`).
    - The full-width-half-maximum (:math:`\\Delta f_{HM}`) value of the peak is then
      determined using linear interpolation of the trace, by difference of the :math:`f`
      at the two points where :math:`||\\Gamma(f)|| = 0.5`.
    - Finally, the quality factor is calculated from :math:`Q_0 = f_0 / \\Delta f_{HM}`.

    Parameters
    ----------
    fvals
        Nominal values of the frequencies
    fsigs
        Error values of the frequencies
    gvals
        Not used
    absgvals
        Absolute values of the reflection coefficient, :math:`|\\Gamma(f)|`

    Returns
    -------
    (Q0, f0): tuple[uc.ufloat, uc.ufloat]
        Fitted quality factor and central frequency of the data.
    """

    maxg = absgvals.max()
    ming = absgvals.min()
    absgamma = (absgvals - maxg) / (ming - maxg)
    ai = np.argmax(absgamma)
    lf = np.interp(0.5, absgamma[:ai], fvals[:ai])
    rf = np.interp(0.5, absgamma[ai:][::-1], fvals[ai:][::-1])
    f0 = uc.ufloat(fvals[ai], fsigs[ai])
    Q0 = f0 / (uc.ufloat(rf, fsigs[ai]) - uc.ufloat(lf, fsigs[ai]))
    return Q0, f0


def _lorentz(x, a, x0, gam, c):
    return a * (gam ** 2 / ((x - x0) ** 2 + gam ** 2)) + c


def lorentz(
    fvals: np.ndarray,
    fsigs: np.ndarray,
    gvals: np.ndarray,
    absgvals: np.ndarray,
    **kwargs: dict
) -> tuple[uc.ufloat, uc.ufloat]:
    """
    Lorentz fit.

    Fits a Lorentz function to the pruned data. The :math:`f_0 = x_0`, and the
    :math:`Q_0` is calculated from :math:`x_0 / \\Delta x = x_0 / (2\\gamma)`.

    - This fit starts by fitting a Lorentzian curve in the form
      :math:`L(x) = a \\frac{\\gamma^2}{(x - x_0)^2 + \\gamma^2} + c` to the
      :math:`|\\Gamma(f)|` trace, obtaining four parameters (:math:`a, x_0, \\gamma, c`).
    - The central frequency :math:`f_0` is simply :math:`x_0`.
    - The width of the Lorentzian is given by :math:`\\Delta x = 2\\gamma`.
    - The quality factor is determined using :math:`Q_0 = x_0 / \\Delta x`
    - Uncertainties of :math:`f_0` and :math:`Q_0` are calculated using the
      covariance matrix of the fit of :math:`L(x)` to :math:`|\\Gamma(f)|`.


    Parameters
    ----------
    fvals
        Nominal values of the frequencies
    fsigs
        Not used
    gvals
        Not used
    absgvals
        Absolute values of the reflection coefficient, :math:`|\\Gamma(f)|`

    Returns
    -------
    (Q0, f0): tuple[uc.ufloat, uc.ufloat]
        Fitted quality factor and central frequency of the data.
    """
    popt, pcov = curve_fit(
        _lorentz,
        fvals,
        absgvals,
        sigma=absgvals,
        absolute_sigma=True,
        p0=[-0.5, fvals[np.argmin(absgvals)], 1e5, 1],
    )
    perr = np.sqrt(np.diag(pcov))
    x0 = uc.ufloat(popt[1], perr[1])
    gam = uc.ufloat(popt[2], perr[2])

    return x0 / (2 * gam), x0


def kajfez(
    fvals: np.ndarray,
    fsigs: np.ndarray,
    gvals: np.ndarray,
    absgvals: np.ndarray,
    **kwargs: dict
) -> tuple[uc.ufloat, uc.ufloat]:
    """
    Kajfez's circle-fitting program.

    Adapted from Q0REFL.m, which is a part of Kajfez, D.: "Linear fractional curve
    fitting for measurement of high Q factors", IEEE Trans. Microwave Theory Techn.
    42 (1994) 1149-1153.

    This fitting process attempts to fit a circle to a near-circular section of
    points on a Smith's chart. It's robust, quick, and reliable, and produces
    reasonable error estimates.

    Parameters
    ----------
    fvals
        Nominal values of the frequencies
    fsigs
        Error values of the frequencies
    gvals
        Complex reflection coefficient values, :math:`\\Gamma(f)`.
    absgvals
        Absolute values of the reflection coefficient, :math:`|\\Gamma(f)|`

    Returns
    -------
    (Q0, f0): tuple[uc.ufloat, uc.ufloat]
        Fitted quality factor and central frequency of the data.
    """
    niter = kwargs.get("iterations", 5)
    fre = fvals
    n = fvals.size
    gam1 = gvals
    agam1 = absgvals
    dia = agam1.min()
    idia = np.argmax(agam1)
    f0 = fre[idia]
    ena = np.ones(n)
    sig = np.ones(3)
    eps = np.ones(n)
    p = np.zeros(n)
    for ite in range(niter):
        x = 2 * (fre / f0 - ena)
        e3 = gam1 * x
        e2 = -ena
        e1 = -x
        F = -gam1
        x2 = abs(x) ** 2
        ga2 = abs(gam1) ** 2
        if ite == 0:
            p = 1.0 / (x2 * (ena + ga2) + ena)
        else:
            p = sig[1] ** 2 / (
                x2 * sig[0] ** 2 + sig[1] ** 2 + (x2 * ga2) * sig[2] ** 2
            )
        E = np.array([e1, e2, e3]).T
        PE = np.array([p * e1, p * e2, p * e3]).T
        C = E.conj().T @ PE  # @?
        q1 = E.conj().T @ (p * F)  # @ ?
        D = np.linalg.inv(C)
        g = D @ q1
        eps = g[0] * e1 + g[1] * e2 + g[2] * e3 - F
        S1sq = eps.conj().T @ (p * eps)
        Fsq = F.conj().T @ (p * F)
        sumden = C[0, 0] * D[0, 0] + C[1, 1] * D[1, 1] + C[2, 2] * D[2, 2]
        for m in range(3):
            sig[m] = np.sqrt(abs(D[m, m] * S1sq / sumden))
        diam1 = 2 * abs(g[1] * g[2] - g[0]) / abs(g[2].conj() - g[2])
        gamc1 = (g[2].conj() * g[1] - g[0]) / (g[2].conj() - g[2])
        gamd1 = g[0] / g[2]
        gam1L2 = 2 * gamc1 - gamd1
        delt = (g[1] - gam1L2) / (gam1L2 * g[2] - g[0])
        f0 = f0 * (1 + 0.5 * delt.real)
    f011 = f0
    dia1 = 2 * abs(g[1] * g[2] - g[0]) / abs(g[2].conj() - g[2])
    flin = np.linspace(fre[0], fre[-1], 201)
    ft = 2 * (flin / f011 - 1)
    gam1com = (g[0] * ft.T + g[1]) / (g[2] * ft.T + 1)
    gam1d = g[0] / g[2]
    gam1c = (g[2].conj() * g[1] - g[0]) / (g[2].conj() - g[2])
    QL1 = g[2].imag

    # QL1 -> Q01
    ang1 = -np.arctan2((-gam1d).imag, (-gam1d).real)
    ang2 = np.arctan2((gam1c - gam1d).imag, (gam1c - gam1d).real)
    angtot = ang1 + ang2
    cob = np.cos(angtot)
    rr2 = (1 - abs(gam1d) ** 2) * 0.5 / (1 - abs(gam1d) * cob)
    dr2 = 2 * rr2
    rr1 = dia1 * 0.5
    coupls = (1 / rr2 - 1) / (1 / rr1 - 1 / rr2)
    rs = 2 / dr2 - 1
    kapa1 = rr1 / (rr2 - rr1)
    Q01 = QL1 * (1 + kapa1)

    # standard deviations
    sdQL1 = np.sqrt((sig[2].real) ** 2 + sig[2] ** 2)
    sddia1an = np.sqrt(
        sig[0] ** 2 * abs(g[2]) ** (-2)
        + sig[1] ** 2
        + abs(g[0] / (g[2] ** 2)) ** 2 * sig[2] ** 2
    )
    equi = abs(gam1 - gamc1)
    avequi = equi.mean()
    sdequi = equi.std()
    sddia1st = sdequi * 2
    sddia1 = sddia1an
    sdkapa1 = sddia1 / (2 - dia1) ** 2
    sdQ01 = np.sqrt((1 + kapa1) ** 2 * sdQL1 ** 2 + QL1 ** 2 * sdkapa1 ** 2)

    return uc.ufloat(Q01, sdQ01), uc.ufloat(f011, fsigs[idia])



"""
File parser for Fusion json data format (json).

This is a fairly detailed data format, including the traces, the calibration applied,
and also the integrated peak areas. If the peak areas are present, this is returned
in the list of timesteps as a ``"peaks"`` entry.

.. note ::
    The detectors in the trace data are not necessarily in a consistent order, which
    may change between different files. Hence, the keys are sorted.

Exposed metadata:
`````````````````

.. code-block:: yaml

    params:
      method:   !!str
      sampleid: !!str
      username: None
      version:  !!str
      valve:    !!int
      datafile: !!str

.. codeauthor:: Peter Kraus
"""
import json
import numpy as np

import yadg.dgutils
from yadg.dgutils.dateutils import str_to_uts


def process(fn: str, encoding: str, timezone: str) -> tuple[list, dict, dict]:
    """
    Fusion json format.

    One chromatogram per file with multiple traces, and pre-analysed results.
    Only a subset of the metadata is retained, including the method name,
    detector names, and information about assigned peaks.

    Parameters
    ----------
    fn
        Filename to process.

    encoding
        Encoding used to open the file.

    timezone
        Timezone information. This should be ``"localtime"``.

    Returns
    -------
    ([chrom], metadata): tuple[list, dict]
        Standard timesteps & metadata tuple.
    """

    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        jsdata = json.load(infile)
    metadata = {
        "filetype": "fusion.json",
        "params": {
            "method": jsdata.get("methodName", "n/a"),
            "sampleid": jsdata.get("annotations", {}).get("name", None),
            "valve": jsdata.get("annotations", {}).get("valcoPosition", None),
            "version": jsdata.get("softwareVersion", {}).get("version", None),
            "datafile": jsdata.get("sequence", {}).get("location", None),
            "username": None,
        },
    }
    chrom = {"fn": str(fn), "traces": {}}
    chrom["uts"] = str_to_uts(jsdata["runTimeStamp"], timezone=timezone)
    detid = 0
    # prepare analysis dictionary:
    an = {"height": {}, "area": {}, "concentration": {}, "xout": {}}
    has_an = False

    # sort detector keys to ensure alphabetic order for ID matching
    for detname in sorted(jsdata["detectors"].keys()):
        detdict = jsdata["detectors"][detname]
        trace = {"id": detid}
        detid += 1
        xmul = detdict["nValuesPerSecond"]
        npoints = detdict["nValuesExpected"]
        assert (
            len(detdict["values"]) == npoints
        ), f"fusion: Inconsistent trace length in file {fn}."
        xsn = np.arange(npoints) / xmul
        xss = np.ones(npoints) / xmul
        xs = [xsn, xss]
        ysn = np.array(detdict["values"])
        yss = np.ones(npoints)
        ys = [ysn, yss]
        trace["t"] = {
            "n": xsn.tolist(),
            "s": xss.tolist(),
            "u": "s",
        }
        trace["y"] = {
            "n": ysn.tolist(),
            "s": yss.tolist(),
            "u": " ",
        }
        trace["data"] = [xs, ys]
        if "analysis" in detdict:
            has_an = True
            for peak in detdict["analysis"]["peaks"]:
                if "label" not in peak:
                    continue
                if "baselinePoints" in peak:
                    nbp = (
                        peak["baselinePoints"]["start"] - peak["baselinePoints"]["end"]
                    )
                else:
                    nbp = 0
                h = {"n": float(peak.get("height", 0.0)), "s": 0.0, "u": " "}
                A = {"n": float(peak.get("area", 0.0)), "s": 0.0, "u": " "}
                c = {"n": float(peak.get("concentration", 0.0)), "s": 0.0, "u": " "}
                x = {
                    "n": float(peak.get("normalizedConcentration", 0.0)) / 100.0,
                    "s": 0.0,
                    "u": " ",
                }
                an["height"][peak["label"]] = h
                an["area"][peak["label"]] = A
                an["concentration"][peak["label"]] = c
                an["xout"][peak["label"]] = x
        chrom["traces"][detname] = trace

    if has_an:
        chrom.update(an)
    return [chrom], metadata
